﻿
Namespace ECC_LIB

    Public Interface ECC_Engine_Interface
        Function ReadData(ByRef flash_page() As Byte, page_size As UInt16, ByRef flash_spare() As Byte, spare_size As UInt16) As ECC_DECODE_RESULT
        Sub WriteData(ByRef flash_page() As Byte, page_size As UInt16, ByRef flash_spare() As Byte, spare_size As UInt16)
    End Interface

    Public Enum ECC_DECODE_RESULT
        NoErrors 'all bits and parity match
        Correctable 'one or more bits dont match but was corrected
        EccError 'ECC has an error (and has been fixed)
        DataEccError 'Both the data and the ECC has an error (only supported on 2-bit+ correction)
        Uncorractable 'more errors than are correctable
        Unknown 'User sent data that was not in 512 byte segments
    End Enum

    Public Enum ecc_algorithum As Integer
        hamming = 0
        reedsolomon = 1
        bch = 2
        dhara = 3
    End Enum

    Public Class ECC_Configuration_Entry
        Public Property Algorithm As ecc_algorithum 'Hamming / RS / BCH
        Public Property Selected As Boolean
        Public Property ChunkSize As UInt16 = 512 'Number of bytes per ECC correction
        Public Property BitError As Byte 'Number of bits that can be corrected per chunk
        Public Property RsSymSize As Byte 'Number of bits per symbol (RS only)
        Public Property InitialOffset As UInt16 'Initial start address in OOB to store ECC codes
        Public Property Spacing As UInt16 'Number of bytes between ECC codes
        Public Property ECC_Reverse As Boolean = False  'Reverses ECC byte order

        Sub New()

        End Sub

        'Returns the total amount of bytes needed to save this profile
        Public Function GetTotalSize(page_size As UInt16) As Integer
            Dim ecc_chunks As Integer = CInt(Math.Ceiling(page_size / ChunkSize))
            Dim ecc_spacing As Integer = ((ecc_chunks - 1) * Me.Spacing)
            Dim ecc_size As Integer = GetEccSize()
            If ecc_size = -1 Then Return -1
            Return (Me.InitialOffset + ecc_spacing + (ecc_chunks * ecc_size))
        End Function

        Public Function GetEccSize() As Integer
            If Me.Algorithm = ecc_algorithum.hamming Then
                Return 3
            ElseIf Me.Algorithm = ecc_algorithum.reedsolomon Then
                Return RS_ECC.PredictSize(ChunkSize, BitError, RsSymSize)
            ElseIf Me.Algorithm = ecc_algorithum.bch Then
                Return BinaryBCH.PredictSize(ChunkSize, BitError)
            ElseIf Me.Algorithm = ecc_algorithum.dhara Then
                Return DharaBCH.PredictSize(ChunkSize, BitError)
            End If
            Return -1
        End Function

        'Indicates if this profile entry is compatible with this flash space
        Public Function IsValid(flash_page As UInt16, flash_spare As UInt16) As Boolean
            Dim total_size = GetTotalSize(flash_page)
            If total_size = -1 Then Return False
            If (total_size > flash_spare) Then Return False
            Return True
        End Function

    End Class

    Public Module Common
        Public NAND_ECC_CFG() As ECC_Configuration_Entry

        Public Function ECC_GetErrorMessage(err As ECC_DECODE_RESULT) As String
            Select Case err
                Case ECC_DECODE_RESULT.NoErrors
                    Return "No Errors found"
                Case ECC_DECODE_RESULT.Correctable
                    Return "Data Corrected"
                Case ECC_DECODE_RESULT.EccError
                    Return "ECC data error"
                Case ECC_DECODE_RESULT.Uncorractable
                    Return "Too many errors"
                Case ECC_DECODE_RESULT.Unknown
                    Return "Block data was invalid"
                Case Else
                    Return String.Empty
            End Select
        End Function

    End Module


End Namespace






